/************************************************************
* $Id: utils.c,v 1.3 2015/11/09 18:23:50 mechanoid Exp $
*  
*  OpenCL demo
*
*       OS : GNU/Linux 4.2.3-1-ARCH x86_64 
* COMPILER : GNU project C and C++ Compiler  v.5.2.0
*
*   AUTHOR : Evgeny S. Borisov
*
*    http://www.mechanoid.kiev.ua
*  e-mail : opencl@mechanoid.kiev.ua
* 
************************************************************/
#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <stdio.h>
#include <string.h>

#include "utils.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
// печатает информацию об ошибках компиляции ядра
int log_compiler_error(cl_program pro, cl_device_id dev) {
   char   *buildLog = NULL;
   size_t buildLogSize = 0;
   cl_int clStatus;
   
   clStatus= clGetProgramBuildInfo(pro, dev, CL_PROGRAM_BUILD_LOG, buildLogSize, buildLog, &buildLogSize);
   CL_CATCH_ERROR(clStatus,"clGetProgramBuildInfo");

   CATCH_MALLOC_ERROR(buildLog=malloc(sizeof(char)*buildLogSize));
   bzero(buildLog, sizeof(char)*buildLogSize);

   clStatus= clGetProgramBuildInfo(pro, dev, CL_PROGRAM_BUILD_LOG, buildLogSize, buildLog, NULL);
   CL_CATCH_ERROR(clStatus,"clGetProgramBuildInfo");

   fprintf(stderr,"KERNEL BUILD LOG:\n%s\n",buildLog);

   free(buildLog);

   return 0;;
} 

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
int check_time(cl_event	evt) { // оценка времени выполнения
   cl_ulong time_start, time_end;
   double total_time;
   cl_int clStatus;

   // считаем затраченное время
   clStatus= clGetEventProfilingInfo(evt, CL_PROFILING_COMMAND_START, sizeof(time_start), &time_start, NULL);
   CL_CATCH_ERROR( clStatus,"clGetEventProfilingInfo") ;
   clStatus= clGetEventProfilingInfo(evt, CL_PROFILING_COMMAND_END, sizeof(time_end), &time_end, NULL);
   CL_CATCH_ERROR( clStatus,"clGetEventProfilingInfo") ;
   total_time = time_end - time_start;
   fprintf(stdout,"\nexecution time: %.5f ms\n", (total_time*10e-6f) );

   return 0;
} 


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
// расшифровка кода ошибки
const char *get_error_string(cl_int code) {
   switch(code){
	  case CL_SUCCESS: return "CL_SUCCESS";
	  // run-time and JIT compiler errors
	  case CL_DEVICE_NOT_FOUND: return "CL_DEVICE_NOT_FOUND";
	  case CL_DEVICE_NOT_AVAILABLE: return "CL_DEVICE_NOT_AVAILABLE";
	  case CL_COMPILER_NOT_AVAILABLE: return "CL_COMPILER_NOT_AVAILABLE";
	  case CL_MEM_OBJECT_ALLOCATION_FAILURE: return "CL_MEM_OBJECT_ALLOCATION_FAILURE";
	  case CL_OUT_OF_RESOURCES: return "CL_OUT_OF_RESOURCES";
	  case CL_OUT_OF_HOST_MEMORY: return "CL_OUT_OF_HOST_MEMORY";
	  case CL_PROFILING_INFO_NOT_AVAILABLE: return "CL_PROFILING_INFO_NOT_AVAILABLE";
	  case CL_MEM_COPY_OVERLAP: return "CL_MEM_COPY_OVERLAP";
	  case CL_IMAGE_FORMAT_MISMATCH: return "CL_IMAGE_FORMAT_MISMATCH";
	  case CL_IMAGE_FORMAT_NOT_SUPPORTED: return "CL_IMAGE_FORMAT_NOT_SUPPORTED";
	  case CL_BUILD_PROGRAM_FAILURE: return "CL_BUILD_PROGRAM_FAILURE";
	  case CL_MAP_FAILURE: return "CL_MAP_FAILURE";
	  case CL_MISALIGNED_SUB_BUFFER_OFFSET: return "CL_MISALIGNED_SUB_BUFFER_OFFSET";
	  case CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST: return "CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST";
	  case CL_COMPILE_PROGRAM_FAILURE: return "CL_COMPILE_PROGRAM_FAILURE";
	  case CL_LINKER_NOT_AVAILABLE: return "CL_LINKER_NOT_AVAILABLE";
	  case CL_LINK_PROGRAM_FAILURE: return "CL_LINK_PROGRAM_FAILURE";
	  case CL_DEVICE_PARTITION_FAILED: return "CL_DEVICE_PARTITION_FAILED";
	  case CL_KERNEL_ARG_INFO_NOT_AVAILABLE: return "CL_KERNEL_ARG_INFO_NOT_AVAILABLE";

	  // compile-time errors
	  case CL_INVALID_VALUE: return "CL_INVALID_VALUE";
	  case CL_INVALID_DEVICE_TYPE: return "CL_INVALID_DEVICE_TYPE";
	  case CL_INVALID_PLATFORM: return "CL_INVALID_PLATFORM";
	  case CL_INVALID_DEVICE: return "CL_INVALID_DEVICE";
	  case CL_INVALID_CONTEXT: return "CL_INVALID_CONTEXT";
	  case CL_INVALID_QUEUE_PROPERTIES: return "CL_INVALID_QUEUE_PROPERTIES";
	  case CL_INVALID_COMMAND_QUEUE: return "CL_INVALID_COMMAND_QUEUE";
	  case CL_INVALID_HOST_PTR: return "CL_INVALID_HOST_PTR";
	  case CL_INVALID_MEM_OBJECT: return "CL_INVALID_MEM_OBJECT";
	  case CL_INVALID_IMAGE_FORMAT_DESCRIPTOR: return "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR";
	  case CL_INVALID_IMAGE_SIZE: return "CL_INVALID_IMAGE_SIZE";
	  case CL_INVALID_SAMPLER: return "CL_INVALID_SAMPLER";
	  case CL_INVALID_BINARY: return "CL_INVALID_BINARY";
	  case CL_INVALID_BUILD_OPTIONS: return "CL_INVALID_BUILD_OPTIONS";
	  case CL_INVALID_PROGRAM: return "CL_INVALID_PROGRAM";
	  case CL_INVALID_PROGRAM_EXECUTABLE: return "CL_INVALID_PROGRAM_EXECUTABLE";
	  case CL_INVALID_KERNEL_NAME: return "CL_INVALID_KERNEL_NAME";
	  case CL_INVALID_KERNEL_DEFINITION: return "CL_INVALID_KERNEL_DEFINITION";
	  case CL_INVALID_KERNEL: return "CL_INVALID_KERNEL";
	  case CL_INVALID_ARG_INDEX: return "CL_INVALID_ARG_INDEX";
	  case CL_INVALID_ARG_VALUE: return "CL_INVALID_ARG_VALUE";
	  case CL_INVALID_ARG_SIZE: return "CL_INVALID_ARG_SIZE";
	  case CL_INVALID_KERNEL_ARGS: return "CL_INVALID_KERNEL_ARGS";
	  case CL_INVALID_WORK_DIMENSION: return "CL_INVALID_WORK_DIMENSION";
	  case CL_INVALID_WORK_GROUP_SIZE: return "CL_INVALID_WORK_GROUP_SIZE";
	  case CL_INVALID_WORK_ITEM_SIZE: return "CL_INVALID_WORK_ITEM_SIZE";
	  case CL_INVALID_GLOBAL_OFFSET: return "CL_INVALID_GLOBAL_OFFSET";
	  case CL_INVALID_EVENT_WAIT_LIST: return "CL_INVALID_EVENT_WAIT_LIST";
	  case CL_INVALID_EVENT: return "CL_INVALID_EVENT";
	  case CL_INVALID_OPERATION: return "CL_INVALID_OPERATION";
	  case CL_INVALID_GL_OBJECT: return "CL_INVALID_GL_OBJECT";
	  case CL_INVALID_BUFFER_SIZE: return "CL_INVALID_BUFFER_SIZE";
	  case CL_INVALID_MIP_LEVEL: return "CL_INVALID_MIP_LEVEL";
	  case CL_INVALID_GLOBAL_WORK_SIZE: return "CL_INVALID_GLOBAL_WORK_SIZE";
	  case CL_INVALID_PROPERTY: return "CL_INVALID_PROPERTY";
	  case CL_INVALID_IMAGE_DESCRIPTOR: return "CL_INVALID_IMAGE_DESCRIPTOR";
	  case CL_INVALID_COMPILER_OPTIONS: return "CL_INVALID_COMPILER_OPTIONS";
	  case CL_INVALID_LINKER_OPTIONS: return "CL_INVALID_LINKER_OPTIONS";
	  case CL_INVALID_DEVICE_PARTITION_COUNT: return "CL_INVALID_DEVICE_PARTITION_COUNT";

	  /* extension errors
	  case CL_INVALID_GL_SHAREGROUP_REFERENCE_KHR: return "CL_INVALID_GL_SHAREGROUP_REFERENCE_KHR";
	  case CL_PLATFORM_NOT_FOUND_KHR: return "CL_PLATFORM_NOT_FOUND_KHR";
	  case CL_INVALID_D3D10_DEVICE_KHR: return "CL_INVALID_D3D10_DEVICE_KHR";
	  case CL_INVALID_D3D10_RESOURCE_KHR: return "CL_INVALID_D3D10_RESOURCE_KHR";
	  case CL_D3D10_RESOURCE_ALREADY_ACQUIRED_KHR: return "CL_D3D10_RESOURCE_ALREADY_ACQUIRED_KHR";
	  case CL_D3D10_RESOURCE_NOT_ACQUIRED_KHR: return "CL_D3D10_RESOURCE_NOT_ACQUIRED_KHR";
	  */
   }
   return "Unknown OpenCL error code";
}


