/************************************************************
* $Id: gemm2.cl,v 1.3 2015/10/31 11:29:57 mechanoid Exp mechanoid $
*  
* OpenCL matrix multiplication 
*
*       OS : GNU/Linux 4.2.3-1-ARCH x86_64 
* COMPILER : GNU project C and C++ Compiler  v.5.2.0
*
*   AUTHOR : Evgeny S. Borisov
*
*    http://www.mechanoid.kiev.ua
*  e-mail : opencl@mechanoid.kiev.ua
* 
************************************************************/
#include "gemm2.h"

// конвертер индексов матрицы в линейный адрес
#define IDX2LIN(i,j,l) ((i)+(j)*(l))

// A[M*K]   B[K*N]   C[M*N]

__kernel void myGEMM2(const int M, const int N, const int K,
					const __global float* A, const __global float* B, __global float* C) 
{
   // 2D номер треда в группе
   const int r = get_local_id(0); 
   const int c = get_local_id(1); 

   // номер ячейки в матрице результата
   const int gr = get_group_id(0)*TS + r; // 0..M
   const int gc = get_group_id(1)*TS + c; // 0..N

   // общий кэш для тредов группы 
   __local float Asub[TS][TS];
   __local float Bsub[TS][TS];

   float acc = 0.0f; // результат работы треда

   for (int t=0; t<K/TS; t++) { // цикл по всем блокам матриц

	  const int tr = t*TS + r;
	  const int tc = t*TS + c;

	  // загружаем блоки в кэш
	  Asub[c][r] = A[ IDX2LIN(gr,tc,M) ];
	  Bsub[c][r] = B[ IDX2LIN(tr,gc,K) ];

	  barrier(CLK_LOCAL_MEM_FENCE); // ждём пока треды группы заполнят общий кэш

	  // вычисляем 
	  for (int k=0; k<TS; k++) { acc += Asub[k][r] * Bsub[c][k]; }

	  barrier(CLK_LOCAL_MEM_FENCE); // ждём пока треды группы завершат вычисления
   }

   C[ IDX2LIN(gr,gc,M) ] = acc; // сохраняем результат
}
