/************************************************************
* $Id: main.c,v 1.4 2015/11/09 18:23:50 mechanoid Exp $
*  
*  OpenCL demo
*
*       OS : GNU/Linux 4.2.3-1-ARCH x86_64 
* COMPILER : GNU project C and C++ Compiler  v.5.2.0
*
*   AUTHOR : Evgeny S. Borisov
*
*    http://www.mechanoid.kiev.ua
*  e-mail : opencl@mechanoid.kiev.ua
* 
************************************************************/
#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <stdio.h>
#include <string.h>

#include "utils.h"
#include "mat_ops_seq.h"

#define STR_BUF_SIZE  1024
#define DEVICE_TYPE  CL_DEVICE_TYPE_ALL

// размеры матриц для обработки
#define N 1024
#define M 512
#define K 256

// матрицы, память host
data_t hA[M*K];
data_t hB[K*N];
data_t hC[M*N];
data_t hC2[M*N];

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
// текст ядра
const char *kernel_src=
#include "kernel.cl.txt"
;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
cl_platform_id	*platforms; // список платформ
cl_device_id	*devices; // список вычислительных устройств

cl_context		context; 
cl_program 		program; // программа для DEVICE
cl_kernel 		kernel; // выполняемое ядро, созданное из программы
cl_event		kernel_event;
cl_command_queue command_queue ; // очередь выполнения для выбранных устройств

int sel_platform_num=0;
int sel_device_num=0;

// матрицы, память device
cl_mem clA; 
cl_mem clB;
cl_mem clC;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
int init_platform();
int init_device();
int make_kernel();
int init_device_mem();
int run_kernel();
int check_result();
int release_resources();

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
int main(int argc, char *argv[]){

   switch(argc){
	  case 2: sel_platform_num=atoi(argv[1]); break;
	  case 3: sel_platform_num=atoi(argv[1]); sel_device_num=atoi(argv[2]); break;
   }
   
   if(init_platform()) return 1;
   if(init_device() ) return 2;
   if(make_kernel()) return 3;
   if(init_device_mem()) return 4;
   run_kernel();
   check_result();
   check_time(kernel_event);
   release_resources();

   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
int init_device_mem() { // выделяем память device для данных
   cl_int clStatus;
   int i;

   clA= clCreateBuffer(context, CL_MEM_READ_ONLY, M*K*sizeof(data_t), NULL, &clStatus);
   CL_CATCH_ERROR(clStatus,"error clCreateBuffer" );

   clB= clCreateBuffer(context, CL_MEM_READ_ONLY, K*N*sizeof(data_t), NULL, &clStatus);
   CL_CATCH_ERROR(clStatus,"error clCreateBuffer" );

   clC= clCreateBuffer(context, CL_MEM_WRITE_ONLY, M*N*sizeof(data_t), NULL, &clStatus);
   CL_CATCH_ERROR(clStatus,"error clCreateBuffer" );

   mat_init_rand(M*K,hA);
   mat_init_rand(K*N,hB);

   bzero(hC,M*N*sizeof(data_t));
   bzero(hC2,M*N*sizeof(data_t));

   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
int run_kernel() {
   cl_int clStatus;
   size_t global[2] = { GLOBAL1, GLOBAL2 };
   size_t local[2] = { LOCAL1, LOCAL2 };
   int a;

   // копируем данные в память DEVICE
   clStatus= clEnqueueWriteBuffer(command_queue, clA, CL_TRUE, 0, M*K*sizeof(data_t), hA, 0, NULL, &kernel_event );
   CL_CATCH_ERROR(clStatus,"clEnqueueWriteBuffer");
   clStatus= clEnqueueWriteBuffer(command_queue, clB, CL_TRUE, 0, K*N*sizeof(data_t), hB, 0, NULL, &kernel_event );
   CL_CATCH_ERROR(clStatus,"clEnqueueWriteBuffer");

   // назначаем параметры ядра
   a=M; clStatus = clSetKernelArg(kernel, 0, sizeof(int), (void*)&a);
   a=N; clStatus |= clSetKernelArg(kernel, 1, sizeof(int), (void*)&a);
   a=K; clStatus |= clSetKernelArg(kernel, 2, sizeof(int), (void*)&a);
   clStatus |= clSetKernelArg(kernel, 3, sizeof(cl_mem), (void*)&clA);
   clStatus |= clSetKernelArg(kernel, 4, sizeof(cl_mem), (void*)&clB);
   clStatus |= clSetKernelArg(kernel, 5, sizeof(cl_mem), (void*)&clC);
   CL_CATCH_ERROR(clStatus,"error clSetKernelArg");

   fprintf(stdout,"run %s kernel, work items: [ %d %d ], workgroup size: [ %d %d ]\n", KERNEL_NAME, global[0], global[1], local[0], local[1] ); 
   
   // запускаем ядро 
   clStatus= clEnqueueNDRangeKernel(command_queue, kernel, 2, NULL, global, local, 0, NULL, &kernel_event);
   CL_CATCH_ERROR(clStatus,"clEnqueueNDRangeKernel");
   
   // считываем результат
   clStatus= clEnqueueReadBuffer(command_queue, clC, CL_TRUE, 0, M*N*sizeof(data_t), hC, 1, &kernel_event, NULL);
   CL_CATCH_ERROR(clStatus,"clEnqueueReadBuffer");

   // ждём завершения всех операций
   clStatus=  clFinish(command_queue);
   CL_CATCH_ERROR( clStatus,"clFinish") ;

   return 0;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
int check_result() {
   //fprintf(stdout,"check result...\n");
   //mat_mult_seq(M,N,K,hA,hB,hC2);
   //if(!mat_cmp(M,N,hC,hC2) ){ fprintf(stdout,"\nsuccess\n"); }
   //mat_print(M,N,hC);
   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
int make_kernel() {
   cl_int clStatus;

   // создать программу из текста ядра
   program = clCreateProgramWithSource(context, 1, &kernel_src, NULL, &clStatus);
   CL_CATCH_ERROR(clStatus,"error clCreateProgramWithSource" );

   // компилируем программу-ядро
   clStatus = clBuildProgram(program, 1, &devices[sel_device_num], NULL, NULL, NULL);
   if(clStatus != CL_SUCCESS) { 
	  log_compiler_error(program, devices[sel_device_num]);  
	  CL_CATCH_ERROR(clStatus,"error clBuildProgram" );
	  return 8; 
   }

   // создаём объект выполняемого ядра
   kernel = clCreateKernel(program,KERNEL_NAME , &clStatus);
   CL_CATCH_ERROR(clStatus,"error clCreateKernel" );

   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
int release_resources() {  // освобождаем ресурсы
   clReleaseMemObject(clA);
   clReleaseMemObject(clB);
   clReleaseMemObject(clC);
 
   clReleaseKernel(kernel);
   clReleaseProgram(program);

   clReleaseCommandQueue(command_queue);

   clReleaseContext(context);

   free(devices);
   free(platforms);

   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
int init_platform() {
   cl_int clStatus;
   char queryBuffer[STR_BUF_SIZE]; 
   cl_uint queryInt; 
   cl_uint num_platforms ;

   // получить количество доступных платформ 
   clStatus= clGetPlatformIDs(0, NULL, &num_platforms);
   CL_CATCH_ERROR(clStatus,"clGetPlatformIDs");
   if (num_platforms <1 ) { fprintf(stderr,"No OpenCL Platforms Found"); return 1; }
   fprintf(stdout, "found %d platforms\n", num_platforms);

   if( !(sel_platform_num < num_platforms) ) { fprintf(stderr,"sel_platform_num out of range\n"); return 1; }
   fprintf(stdout, "select platform #%d\n", sel_platform_num);

   // создаём буфер для списка платформ
   CATCH_MALLOC_ERROR( platforms = (cl_platform_id *)malloc(num_platforms*sizeof(cl_platform_id)) );
   // читаем список платформ
   clStatus= clGetPlatformIDs (num_platforms, platforms, NULL) ;
   CL_CATCH_ERROR(clStatus,"clGetPlatformIDs");

   // печатаем имя выбранной платформы	
   clStatus=clGetPlatformInfo (platforms[sel_platform_num], CL_PLATFORM_NAME, STR_BUF_SIZE, queryBuffer, NULL);
   CL_CATCH_ERROR(clStatus,"clGetPlatformInfo");
   fprintf(stdout,"CL_PLATFORM_NAME : %s\n", queryBuffer); 

   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
int init_device() {
   cl_int clStatus;
   char queryBuffer[STR_BUF_SIZE]; 
   cl_uint queryInt; 
   cl_uint num_devices;
   int i=0;
   cl_uint wgDim; 
   size_t *wgSz;
   // size_t wgSzMin;

   // получить количество доступных устройств на платформе номер sel_platform_num
   clStatus= clGetDeviceIDs (platforms[sel_platform_num], CL_DEVICE_TYPE_ALL, 0, NULL, &num_devices) ;
   CL_CATCH_ERROR(clStatus,"clGetDeviceIDs");
   if (num_devices<1 ) { fprintf(stderr,"No OpenCL devices found for this platform"); return 1; }
   fprintf(stdout, "found %d devices for this platform\n", num_devices);
   if( !(sel_device_num< num_devices) ) { fprintf(stderr,"sel_device_num out of range\n"); return 1; }
   fprintf(stdout, "select device #%d\n", sel_device_num);

   // создаём буфер для списка устройств 
   CATCH_MALLOC_ERROR( devices=(cl_device_id *)malloc(sizeof(cl_device_id) * num_devices) );

   // получить список доступных устройств типа DEVICE_TYPE на платформе номер sel_platform_num
   clStatus= clGetDeviceIDs (platforms[sel_platform_num], DEVICE_TYPE, num_devices, devices, &num_devices) ;
   CL_CATCH_ERROR(clStatus,"clGetDeviceIDs");

   // печатаем имя выбранного устройства
   clStatus=clGetDeviceInfo(devices[sel_device_num], CL_DEVICE_NAME, STR_BUF_SIZE, queryBuffer, NULL);
   CL_CATCH_ERROR(clStatus,"clGetDeviceInfo");
   fprintf(stdout,"CL_DEVICE_NAME: %s\n", queryBuffer); 

   // печатаем количество вычислительных модулей устройства
   clStatus=clGetDeviceInfo(devices[sel_device_num], CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(queryInt), &queryInt, NULL);
   CL_CATCH_ERROR(clStatus,"clGetDeviceInfo");
   fprintf(stdout,"CL_DEVICE_MAX_COMPUTE_UNITS: %d\n", queryInt); 
   
   // получаем максимальную размерность WORKGROUP
   clStatus=clGetDeviceInfo(devices[sel_device_num], CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS, sizeof(wgDim), &wgDim, NULL);
   CL_CATCH_ERROR(clStatus,"clGetDeviceInfo");
   // fprintf(stdout,"CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS: %d\n", wgDim); 
   
   // печатаем максимальные размеры WORKGROUP
   CATCH_MALLOC_ERROR( wgSz=(size_t*)malloc(sizeof(size_t) * wgDim) );
   clStatus=clGetDeviceInfo(devices[sel_device_num], CL_DEVICE_MAX_WORK_ITEM_SIZES, sizeof(size_t)*wgDim,wgSz, NULL);
   fprintf(stdout,"CL_DEVICE_MAX_WORK_ITEM_SIZES: [");
   for(i=0;i<wgDim;i++){ fprintf(stdout," %d ",wgSz[i]); }
   fprintf(stdout,"]\n");

    free(wgSz);

   // создаём контекст для устройства sel_device_num
   context = clCreateContext( NULL, 1, &devices[sel_device_num], NULL, NULL, &clStatus);
   CL_CATCH_ERROR(clStatus,"error clCreateContext");

   // создаём очередь выполнения 
   // command_queue = clCreateCommandQueueWithProperties(context, devices[sel_device_num], NULL, &clStatus);
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
   // для совместимости с nVidia SDK
   command_queue = clCreateCommandQueue(context, devices[sel_device_num], CL_QUEUE_PROFILING_ENABLE, &clStatus); 
#pragma GCC diagnostic pop
   CL_CATCH_ERROR(clStatus,"error clCreateCommandQueue" );

   return 0;

}


